function FourierShammasSeries2(fx,gx,xRange,Lb,Ub,runNum,sFxName,diaryFilename)
% FourierShammasSeries2 implements the Fourier-Shammas Cosine Series
% Model:
% y = a(0) + a(1)*cos(C(1)*gx(1,A1,B1)*x+Os1) +...+ a(n)*cos(C(n)*gx(n,An,Bn)*x+Osn)
clc
global bDeleteIfExists
global bUseDiary
global xdata
global ydata
global order  
global ggx

warning('off','all')
if isempty(sFxName)
  sFxName = getFuncName(fx);
end
xdata = xRange';
ydata = xdata;
for i=1:length(xdata)
  ydata(i)=fx(xdata(i));
end
ggx = gx;

fprintf('Fitting %s in range (%f, %f)\n', sFxName, min(xdata),max(xdata));

options = optimoptions('particleswarm', 'Display', 'iter');
[x,psAICc] = particleswarm(@optimFunc,length(Lb),Lb,Ub,options);
if isinf(psAICc), psAICc = -1e+99; end
C = zeros(order,1);
Ac = zeros(order,1);
Bc = zeros(order,1);
Oc = zeros(order,1);
C = x(1:order);
Ac = x(order+1:2*order);
Bc = x(2*order+1:3*order);
Oc = x(3*order+1:4*order);

if bUseDiary
  diaryFilename = strrep(diaryFilename, ".txt", strcat("_", num2str(order),"_cos_run", num2str(runNum),".txt"));
  if exist(diaryFilename, 'file')==2
    if bDeleteIfExists
      delete(diaryFilename);
    else
      return;
    end
  end
end
X = [];
for i=1:order
  g = C(i)*gx(i,Ac(i),Bc(i));
  xs = cos(xdata.*g + Oc(i));
  X = [X;xs']; 
end
X = X';
lm = fitlm(X,ydata);
if bUseDiary
  diary(diaryFilename)
end
fprintf('Fitting %s in range (%f, %f)\n', sFxName, min(xdata),max(xdata));
sp = getFuncName(gx);
fprintf('Fourier Shammas Series factor is %s\n', sp);
format long
disp(lm);
anva = anova(lm,'summary');
disp(anva);
format short
fprintf('Model is ----------------------------------\n')
fprintf("y = %e +\n", lm.Coefficients{1,1});
k=2;
for i=1:order
  g = C(i)*gx(i,Ac(i),Bc(i));
  if g >= 0 && Oc(i) > 0
    fprintf("    %e * cos(%e * x + %f) + \n", lm.Coefficients{k,1}, g, Oc(i));
  elseif g >= 0 && Oc(i) < 0
    fprintf("    %e * cos(%e * x - %f) + \n", lm.Coefficients{k,1}, g, abs(Oc(i)));
  elseif g >= 0 && Oc(i) == 0
    fprintf("    %e * cos(%e * x) + \n", lm.Coefficients{k,1}, g);
  elseif g < 0 && Oc(i) > 0
    fprintf("    %e * cos((%e) * x + %f) + \n", lm.Coefficients{k,1}, g, Oc(i));
  elseif g < 0 && Oc(i) == 0
    fprintf("    %e * cos((%e) * x) + \n", lm.Coefficients{k,1}, g);
  else % both g and Oc(i) are negative
    fprintf("    %e * cos((%e) * x - %f) + \n", lm.Coefficients{k,1}, g, abs(Oc(i)));
  end
  k = k + 1;
end
fprintf("\n");

lstFactor = [];
lstOffset = [];
for i=1:order
  g = C(i)*gx(i,Ac(i),Bc(i));
  lstFactor = [lstFactor,g];
  lstOffset = [lstOffset, Oc(i)];
end
fprintf('List of factors: [');
for i=1:order-1
 fprintf('%f, ', lstFactor(i));
end
fprintf('%f]\n', lstFactor(order));
fprintf('List of offsets: [');
for i=1:order-1
 fprintf('%f, ', lstOffset(i));
end
fprintf('%f]\n', lstOffset(order));
fprintf('Fitting %s in range (%f, %f)\n', sFxName, min(xdata),max(xdata));
n = length(xdata);
sumsqr = 0;
for i=1:n
  yc = lm.Coefficients{1,1};
  for j=2:order+1 % length(lm.Coefficients{:,1})
    yc = yc + lm.Coefficients{j,1} * cos(xdata(i)*lstFactor(j-1)+Oc(j-1));
  end
  sumsqr = sumsqr + (ydata(i) - yc)^2;
end  
k = order + 1;
fprintf('MSS of errors squared = %e\n', sqrt(sumsqr)/n);
fprintf("R-Squared = %12.8f\n", lm.Rsquared.Ordinary);
fprintf("R-Squared Adjusted = %12.8f\n", lm.Rsquared.Adjusted);
AIC = lm.ModelCriterion.AIC;
AICc = AIC + 2*k*(1 + (k+1)/(n-k-1));
fprintf('Particle swarm AICc =  %e\n', psAICc);
fprintf('AIC =  %e\n', AIC);
fprintf('AICc =  %e\n', AICc);

if bUseDiary
  diary off
end
end

function AICc = optimFunc(x)
  global xdata
  global ydata
  global order
  global ggx
  
  C = zeros(order,1);
  Ac = zeros(order,1);
  Bc = zeros(order,1);
  Oc = zeros(order,1);
  C = x(1:order);
  Ac = x(order+1:2*order);
  Bc = x(2*order+1:3*order);
  Oc = x(3*order+1:4*order);
  X = [];
  for i=1:order
  	g = C(i)*ggx(i,Ac(i),Bc(i));
    xs = cos(xdata*g + Oc(i));
    X = [X;xs']; 
  end
  X = X';
  lm = fitlm(X,ydata);

  n = length(xdata);
  k = order + 1;
  AIC = lm.ModelCriterion.AIC;
  AICc = AIC + 2*k*(1 + (k+1)/(n-k-1));
  if isinf(AICc), AICc = -1e+99; end
end

function sFx = getFuncName(fx)
  sFx = func2str(fx);
  if sFx(1:2)=="@("
    i = strfind(sFx,")");
    sFx = sFx(i(1)+1:end);
  elseif sFx(1)=="@"
    sFx = strcat(sFx(2:end),".m");
  else
    % return sFx as is
  end
end